__VERSION__ = "0.1.0b1"

import enum


TEXT_UNKNOWN = "Unknown"
TEXT_OPEN = "Open"
TEXT_CLOSED = "Closed"
TEXT_STATUS_HOME = "HOME"
TEXT_STATUS_AWAY = "AWAY"
TEXT_STATUS_DISARM = "DISARM"


class RequestType:
    GET = "GET"
    POST = "POST"


class VisonicURL:
    BASE = "https://{}/rest_api"  #'https://[hostname]/rest_api/[rest_version]

    ACCESS_GRANT = "access/grant"
    ACCESS_REVOKE = "access/revoke"
    ACTIVATE_SIREN = "activate_siren"
    ALARMS = "alarms"
    ALERTS = "alerts"
    APP_TYPE = "apptype"
    AUTH = "auth"
    CAMERAS = "cameras"
    DEVICES = "devices"
    DISABLE_SIREN = "disable_siren"
    EVENTS = "events"
    FEATURE_SET = "feature_set"
    HOME_AUTOMATION_DEVICES = "home_automation_devices"
    LOCATIONS = "locations"
    MAKE_VIDEO = "make_video"
    NOTIFICATIONS_EMAIL = "notifications/email"
    PANEL_LOGIN = "panel/login"
    PANEL_ADD = "panel/add"
    PANEL_INFO = "panel_info"
    PANEL_RENAME = "panel/rename"
    PANEL_UNLINK = "panel/unlink"
    PANELS = "panels"
    PASSWORD_RESET = "password/reset"
    PASSWORD_RESET_COMPLETE = "password/reset/complete"
    PROCESS_STATUS = "process_status?process_tokens={}"
    SET_BYPASS_ZONE = "set_bypass_zone"
    SET_NAME = "set_name"
    SET_STATE = "set_state"
    SET_USER_CODE = "set_user_code"
    SMART_DEVICES = "smart_devices"
    SMART_DEVICES_SETTINGS = "smart_devices/settings"
    STATUS = "status"
    TROUBLES = "troubles"
    USERS = "users"
    VERSION = "version"
    WAKEUP_SMS = "wakeup_sms"



from dataclasses import dataclass
from .classes import BaseClass, title_case
from .const import TEXT_CLOSED, TEXT_OPEN, TEXT_UNKNOWN


@dataclass
class Device(BaseClass):
    """Base class definition of a device in the alarm system."""

    # Device properties
    @property
    def bypass(self) -> bool | None:
        return self._get_nested_key("traits.bypass.enabled", None)

    @property
    def device_number(self) -> int:
        return self._data.get("device_number", 0)

    @property
    def device_type(self) -> str:
        return self._data.get("device_type")

    @property
    def enrollment_id(self) -> str:
        return self._data.get("enrollment_id")

    @property
    def id(self) -> int:
        return self._data.get("id", 0)

    @property
    @title_case
    def location(self) -> str:
        return self._get_nested_key("traits.location.name")

    @property
    def name(self) -> str:
        return self._data.get("name", TEXT_UNKNOWN)

    @property
    def partitions(self) -> list:
        return self._data.get("partitions", [])

    @property
    def preenroll(self) -> bool:
        return self._data.get("preenroll", False)

    @property
    def removable(self) -> bool:
        return self._data.get("removable", False)

    @property
    def renamable(self) -> bool:
        return self._data.get("renamable", False)

    @property
    def soak(self) -> bool:
        return self._get_nested_key("traits.soak.enabled", False)

    @property
    def subtype(self) -> str:
        return self._data.get("subtype", TEXT_UNKNOWN)

    @property
    def warnings(self) -> list:
        return self._data.get("warnings", [])

    @property
    def zone_type(self) -> str:
        return self._data.get("zone_type", TEXT_UNKNOWN)


@dataclass
class CameraDevice(Device):
    """Camera device class definition."""


@dataclass
class ContactDevice(Device):
    """Contact device class definition."""

    @property
    def state(self) -> str:
        """Returns the current state of the contact."""
        if not self.warnings:
            return TEXT_CLOSED
        for warning in self.warnings:
            if warning["type"] == TEXT_OPEN:
                return TEXT_OPEN


@dataclass
class MotionDevice(Device):
    """Motion sensor device class definition."""

    @property
    def brightness(self) -> int:
        return self._get_nested_key("traits.meteo_info.brightness.value")

    @property
    def temperature(self) -> float:
        return self._get_nested_key("traits.meteo_info.temperature.value")


@dataclass
class GenericDevice(Device):
    """Smoke device class definition."""


@dataclass
class GSMDevice(Device):
    """GSM device class definition."""

    @property
    def signal_level(self) -> str | None:
        return self._get_nested_key("traits.signal_level.level")


@dataclass
class KeyFobDevice(Device):
    """KeyFob device class definition."""

    @property
    def owner_id(self) -> int:
        return self._get_nested_key("traits.owner.id", 0)

    @property
    def owner_name(self) -> str:
        return self._get_nested_key("traits.owner.name", TEXT_UNKNOWN)


@dataclass
class PGMDevice(Device):
    """PGM device class definition."""

    @property
    def parent_id(self) -> int:
        return self._get_nested_key("traits.parent.id", 0)

    @property
    def parent_port(self) -> int:
        return self._get_nested_key("traits.parent.port", 0)


@dataclass
class SmokeDevice(Device):
    """Smoke device class definition."""



from dataclasses import dataclass
import functools
import inspect

from .const import TEXT_UNKNOWN


# Decorator function to set string output to title case
def title_case(func):
    def wrapper(*args, **kwargs):
        if result := func(*args, **kwargs):
            return str(result).title()
        return result

    return wrapper


@dataclass
class BaseClass:
    """Base class"""

    _data: dict

    def __repr__(self):
        r = ""
        attrs = inspect.getmembers(self, lambda a: not (inspect.isroutine(a)))
        for i, a in enumerate([attr for attr in attrs if self._is_property(attr)]):
            if i:
                r = r + ", "
            r = r + f"{a[0]} = {getattr(self, a[0])}"
        return f"{type(self).__name__}({r})"

    def __str__(self):
        r = {}
        attrs = inspect.getmembers(self, lambda a: not (inspect.isroutine(a)))
        for a in [attr for attr in attrs if self._is_property(attr)]:
            r[a[0]] = getattr(self, a[0])
        return f"{str(type(self))}: {r}"

    def _is_property(self, attr) -> bool:
        if not (
            attr[0].startswith("__")
            and attr[0].endswith("__")
            or attr[0].startswith("_")
        ):
            return True
        return False

    def _get_nested_key(self, path, *default):
        """Get key value in json by dotted notation or return default"""
        try:
            return functools.reduce(lambda x, y: x[y], path.split("."), self._data)
        except KeyError:
            if default:
                return default[0]
            return None


@dataclass
class Camera(BaseClass):
    """Class definition of an event in the alarm system."""

    # Camera properties
    @property
    @title_case
    def location(self) -> str:
        return self._data.get("location")

    @property
    def partitions(self) -> list[int]:
        return self._data.get("partitions", [])

    @property
    def preenroll(self) -> bool:
        return self._data.get("preenroll", False)

    @property
    def preview_path(self) -> str:
        return self._data.get("preview_path")

    @property
    def status(self) -> str:
        return self._data.get("status")

    @property
    def timestamp(self) -> str:
        # TODO: COnvert to datetime
        return self._data.get("timestamp")

    @property
    def zone(self) -> int:
        return self._data.get("zone")

    @property
    @title_case
    def zone_name(self) -> str:
        return self._data.get("zone_name")


@dataclass
class Event(BaseClass):
    """Class definition of an event in the alarm system."""

    # Event properties
    @property
    def id(self) -> int:
        """User ID."""
        return self._data.get("event", 0)

    @property
    def type_id(self) -> int:
        """Event type ID."""
        return self._data.get("type_id", 0)

    @property
    @title_case
    def label(self) -> str:
        """Event label."""
        return self._data.get("label")

    @property
    def description(self) -> str:
        """Event description."""
        return self._data.get("description")

    @property
    @title_case
    def appointment(self) -> str:
        """Event appointment."""
        return self._data.get("appointment")

    @property
    def datetime(self) -> str:
        """Event datetime."""
        # TODO: make datetime
        return self._data.get("datetime")

    @property
    def video(self) -> bool:
        """Event has video."""
        return self._data.get("video", False)

    @property
    @title_case
    def device_type(self) -> str:
        """Event device type."""
        return self._data.get("device_type")

    @property
    def zone(self) -> int:
        """Event zone."""
        return self._data.get("zone")

    @property
    def partitions(self) -> list[int]:
        """Event partitions."""
        return self._data.get("partitions")

    @property
    def name(self) -> str:
        """Event name."""
        return self._data.get("name")


@dataclass
class FeatureSet(BaseClass):
    """Class definition of an event in the alarm system."""

    # Event properties
    @property
    def events_enabled(self):
        return self._get_nested_key("events.is_enabled")

    @property
    def datetime_enabled(self):
        return self._get_nested_key("datetime.is_enabled")

    @property
    def partitions_enabled(self):
        return self._get_nested_key("partitions.is_enabled")

    @property
    def partitions_has_labels(self):
        return self._get_nested_key("partitions.is_labels_enabled")

    @property
    def partitions_max_count(self):
        return self._get_nested_key("partitions.max_partitions")

    @property
    def devices_enabled(self):
        return self._get_nested_key("devices.is_enabled")

    @property
    def sirens_can_enable(self):
        return self._get_nested_key("sirens.can_enable")

    @property
    def sirens_can_disable(self):
        return self._get_nested_key("sirens.can_disable")

    @property
    def home_automation_devices_enabled(self):
        return self._get_nested_key("home_automation_devices.is_enabled")

    @property
    def state_enabled(self):
        return self._get_nested_key("state.is_enabled")

    @property
    def state_can_set(self):
        return self._get_nested_key("state.can_set")

    @property
    def state_can_get(self):
        return self._get_nested_key("state.can_get")

    @property
    def faults_enabled(self):
        return self._get_nested_key("faults.is_enabled")

    @property
    def diagnostic_enabled(self):
        return self._get_nested_key("diagnostic.is_enabled")

    @property
    def wifi_enabled(self):
        return self._get_nested_key("wifi.is_enabled")


@dataclass
class Location(BaseClass):
    """Class definition of a location in the alarm system."""

    # Location properties
    @property
    def id(self):
        """Location ID."""
        return self._data.get("hel_id")

    @property
    @title_case
    def name(self):
        """Location name."""
        return self._data.get("name")

    @property
    def is_editable(self):
        """Location is editable."""
        return self._data.get("is_editable")


@dataclass
class PanelInfoPartition(BaseClass):
    @property
    def id(self) -> int:
        return self._data.get("id")

    @property
    def active(self) -> bool:
        return self._data.get("active")

    @property
    def exit_delay_time(self) -> int:
        return self._data.get("exit_delay_time")

    @property
    def state_set(self) -> str:
        return self._data.get("state_set")

    @property
    def name(self) -> str:
        return self._data.get("name")


@dataclass
class PanelInfoFeatures(BaseClass):
    @property
    def video_on_demand(self) -> bool:
        return self._data.get("video_on_demand")

    @property
    def alert(self) -> bool:
        return self._data.get("alert")

    @property
    def enabling_siren(self) -> bool:
        return self._data.get("enabling_siren")

    @property
    def disabling_siren(self) -> bool:
        return self._data.get("disabling_siren")

    @property
    def wi_fi_connection(self) -> bool:
        return self._data.get("wi_fi_connection")

    @property
    def set_date_time(self) -> bool:
        return self._data.get("set_date_time")

    @property
    def outputs_setup(self) -> bool:
        return self._data.get("outputs_setup")


@dataclass
class PanelInfo(BaseClass):
    """Class definition of the general alarm system information."""

    # PanelInfo properties
    @property
    @title_case
    def bypass_mode(self) -> str:
        """Bypass Mode"""
        return self._data.get("bypass_mode")

    @property
    @title_case
    def current_user(self) -> str:
        """Current User"""
        return self._data.get("current_user")

    @property
    def local_wakeup_needed(self) -> bool:
        """Local Wakeup Needed"""
        return self._data.get("local_wakeup_needed")

    @property
    @title_case
    def manufacturer(self) -> str:
        """Manufacturer"""
        return self._data.get("manufacturer")

    @property
    @title_case
    def model(self) -> str:
        """Model name"""
        return self._data.get("model")

    @property
    def remote_admin_requires_user_acceptance(self) -> bool:
        """Programming requires user acceptance"""
        return self._data.get(
            "remote_switch_to_programming_mode_requires_user_acceptance"
        )

    @property
    def serial(self) -> str:
        """Serial no"""
        return self._data.get("serial")

    @property
    def partitions(self) -> list[PanelInfoPartition]:
        """Partitions info"""
        return list(
            [
                PanelInfoPartition(partition)
                for partition in self._data.get("partitions")
            ]
        )

    @property
    def features(self) -> PanelInfoFeatures:
        return PanelInfoFeatures(self._data.get("features"))


@dataclass
class Panel(BaseClass):
    """Class definition of the general alarm system information."""

    # Panel properties
    @property
    def panel_serial(self) -> str:
        return self._data.get("panel_serial")

    @property
    def alias(self) -> str:
        return self._data.get("alias")


@dataclass
class Partition(BaseClass):
    """Class definition of a partition in the alarm system."""

    # Partition properties
    @property
    def id(self) -> int:
        return self._data.get("id")

    @property
    def state(self):
        return self._data.get("state")

    @property
    def status(self) -> str:
        return self._data.get("status")

    @property
    def ready(self) -> bool:
        return self._data.get("ready")

    @property
    def options(self) -> list:
        return self._data.get("options")


@dataclass
class Process(BaseClass):
    """Class definition of a process in the alarm system."""

    # Partition properties
    @property
    def token(self) -> str:
        return self._data.get("token")

    @property
    def status(self) -> str:
        return self._data.get("status")

    @property
    def message(self) -> str:
        return self._data.get("message")

    @property
    def error(self) -> str:
        return self._data.get("error")


@dataclass
class Status(BaseClass):
    """Class definition representing the status of the alarm system."""

    # Status properties
    @property
    def connected(self) -> bool:
        return self._data.get("connected")

    @property
    def bba_connected(self) -> bool:
        return self._get_nested_key("connected_status.bba.is_connected", False)

    @property
    def bba_state(self) -> str:
        return self._get_nested_key("connected_status.bba.state", TEXT_UNKNOWN)

    @property
    def gprs_connected(self) -> bool:
        return self._get_nested_key("connected_status.gprs.is_connected", False)

    @property
    def gprs_state(self) -> str:
        return self._get_nested_key("connected_status.grps.state", TEXT_UNKNOWN)

    @property
    def discovery_completed(self) -> bool:
        return self._get_nested_key("discovery.completed")

    @property
    def discovery_stages(self) -> int:
        return self._get_nested_key("discovery.stages")

    @property
    def discovery_in_queue(self) -> int:
        return self._get_nested_key("discovery.in_queue")

    @property
    def discovery_triggered(self) -> bool:
        return self._get_nested_key("discovery.triggered")

    @property
    def partitions(self) -> list[Partition]:
        return [Partition(partition) for partition in self._data.get("partitions", [])]

    @property
    def rssi_level(self) -> int:
        return self._get_nested_key("rssi.level")

    @property
    def rssi_network(self) -> str:
        return self._get_nested_key("rssi.network")


@dataclass
class Trouble(BaseClass):
    """Class definition of a trouble in the alarm system."""

    # Trouble properties
    @property
    def device_type(self) -> str:
        """Device type."""
        return self._data.get("device_type")

    @property
    @title_case
    def location(self) -> str:
        """Location."""
        return self._data.get("location")

    @property
    def partitions(self) -> list[int]:
        """Partitions."""
        return self._data.get("partitions")

    @property
    def trouble_type(self) -> str:
        """Trouble type."""
        return self._data.get("trouble_type")

    @property
    def zone(self) -> int:
        """Zone ID."""
        return self._data.get("zone")

    @property
    @title_case
    def zone_name(self) -> str:
        """Zone type."""
        return self._data.get("zone_name")

    @property
    def zone_type(self) -> str:
        """Zone type."""
        return self._data.get("zone_type")


@dataclass
class User(BaseClass):
    """Class definition of a user in the alarm system."""

    # User properties
    @property
    def id(self) -> int:
        """User ID."""
        return self._data.get("id")

    @property
    @title_case
    def name(self) -> str:
        """User name."""
        return self._data.get("name")

    @property
    def email(self) -> str:
        """User email."""
        return self._data.get("email")

    @property
    def partitions(self) -> list[int]:
        """Device is active."""
        return self._data.get("partitions")


@dataclass
class WakeupSMS(BaseClass):
    """Class definition of a wakeup SMS in the alarm system."""

    # Wakeup SMS properties
    @property
    def phone_number(self) -> str:
        return self._data.get("phone")

    @property
    def message(self) -> str:
        return self._data.get("sms")



from .classes import (
    Camera,
    Event,
    FeatureSet,
    Location,
    PanelInfo,
    Panel,
    Process,
    Status,
    Trouble,
    User,
    WakeupSMS,
)
from .core import API
from .device_definitions import DEVICE_SUBTYPES, DEVICE_TYPES
from .devices import GenericDevice
from .exceptions import UnsupportedRestAPIVersionError


class Setup(object):
    """Class definition of the main alarm system."""

    def __init__(self, hostname, app_id, api_version="latest"):
        """Initiate the connection to the REST API."""
        self.__api = API(hostname, app_id)
        # self.set_rest_version(api_version)

    # System properties
    @property
    def api(self):
        """Return the API for direct access."""
        return self.__api

    def access_grant(self, user_id, email):
        """Grant a user access to the alarm panel via the API."""
        return self.__api.access_grant(user_id, email)

    def access_revoke(self, user_id):
        """Revoke access to the alarm panel via the API for a user."""
        return self.__api.access_revoke(user_id)

    def activate_siren(self):
        """Activate the siren (sound the alarm)."""
        return self.__api.activate_siren()["process_token"]

    def arm_home(self, partition=-1):
        """Send Arm Home command to the alarm system."""
        return self.__api.arm_home(partition)["process_token"]

    def arm_away(self, partition=-1):
        """Send Arm Away command to the alarm system."""
        return self.__api.arm_away(partition)["process_token"]

    def authenticate(self, email, password):
        """Try to authenticate against the API with an email address and password."""
        self.set_rest_version("latest")
        return self.__api.authenticate(email, password)

    def connected(self):
        """Check if the API server is connected to the alarm panel"""
        return self.get_status().connected

    def disable_siren(self, mode="all"):
        """Disable the siren (mute the alarm)."""
        return self.__api.disable_siren(mode=mode)["process_token"]

    def disarm(self, partition=-1):
        """Send Disarm command to the alarm system."""
        return self.__api.disarm(partition)["process_token"]

    def get_cameras(self):
        """Fetch all the devices that are available."""
        cameras = self.__api.get_cameras()
        return [Camera(camera) for camera in cameras]

    def get_devices(self):
        """Fetch all the devices that are available."""
        device_list = []
        devices = self.__api.get_devices()

        for device in devices:
            if DeviceClass := DEVICE_SUBTYPES.get(device["subtype"]):
                device_list.append(DeviceClass(device))
            elif DeviceClass := DEVICE_TYPES.get(device["device_type"]):
                device_list.append(DeviceClass(device))
            else:
                device_list.append(GenericDevice(device))

        return device_list

    def get_events(self, timestamp_hour_offset=2):
        """Get the last couple of events (60 events on my system)."""
        events = self.__api.get_events()
        return [Event(event) for event in events]

    def get_feature_set(self):
        """Fetch the locations associated with the alarm system."""
        feature_set = self.__api.get_feature_set()
        return FeatureSet(feature_set)

    def get_locations(self):
        """Fetch the locations associated with the alarm system."""
        locations = self.__api.get_locations()
        return [Location(location) for location in locations]

    def get_panel_info(self):
        """Fetch basic information about the alarm system."""
        gpi = self.__api.get_panel_info()
        return PanelInfo(gpi)

    def get_panels(self):
        """Fetch a list of panels associated with the user."""
        panels = self.__api.get_panels()
        return [Panel(panel) for panel in panels]

    def get_process_status(self, process_token):
        """Fetch the status information associated with a process token."""
        processes = self.__api.get_process_status(process_token)
        return [Process(process) for process in processes]

    def get_rest_versions(self):
        """Fetch the supported API versions."""
        return self.api.get_version_info()["rest_versions"]

    def get_status(self):
        """Fetch the current state of the alarm system."""

        status = self.__api.get_status()
        return Status(status)

    def get_troubles(self):
        """Fetch all the troubles that are available."""
        troubles = self.__api.get_troubles()
        return [Trouble(trouble) for trouble in troubles]

    def get_users(self):
        """Fetch a list of users in the alarm system."""
        users = self.__api.get_users()
        return [User(user) for user in users["users"]]

    def get_wakeup_sms(self):
        """Fetch a list of users in the alarm system."""
        wakeup_sms = self.__api.get_wakeup_sms()
        return WakeupSMS(wakeup_sms)

    def panel_add(self, alias, panel_serial, master_user_code, access_proof=None):
        """Add a new alarm panel to the user account. A master user code is required."""
        return self.__api.panel_add(alias, panel_serial, access_proof, master_user_code)

    def panel_login(self, panel_serial, user_code):
        """Establish a connection between the alarm panel and the API server."""
        return self.__api.panel_login(panel_serial, user_code)

    def panel_rename(self, alias, panel_serial):
        """Rename an alarm panel."""
        return self.__api.panel_rename(alias, panel_serial)

    def panel_unlink(self, panel_serial, password, app_id):
        """Unlink an alarm panel from the user account."""
        return self.__api.panel_unlink(panel_serial, password, app_id)

    def password_reset(self, email):
        """Send a password reset link to the email address provided in the email argument."""
        return self.__api.password_reset(email)

    def password_reset_complete(self, reset_password_code, new_password):
        """Complete the password reset by entering the reset code received in the email and a new password."""
        return self.__api.password_reset_complete(reset_password_code, new_password)[
            "user_token"
        ]

    def set_bypass_zone(self, zone, set_enabled):
        """Enabled or disable zone bypassing (for example, bypass a sensor to disable it)."""
        return self.__api.set_bypass_zone(zone, set_enabled)["process_token"]

    def set_name_user(self, user_id, name):
        """Set the name of a user by user ID."""
        return self.__api.set_name("USER", user_id, name)["process_token"]

    def set_rest_version(self, version="latest"):
        """
        Fetch the supported versions from the API server and automatically
        configure the library to use the latest version supported by the server,
        unless overridden in the version parameter.
        """
        rest_versions = self.api.get_version_info()["rest_versions"]
        rest_versions.sort(key=float)
        if version == "latest":
            self.__api.set_rest_version(rest_versions[-1])
        elif version in rest_versions:
            self.__api.set_rest_version(version)
        else:
            raise UnsupportedRestAPIVersionError(
                f"Rest API version {version} is not supported by server."
            )

    def set_user_code(self, user_id, user_code):
        """Set the code of a user by user ID."""
        return self.__api.set_user_code(user_code, user_id)["process_token"]



from .devices import (
    CameraDevice,
    ContactDevice,
    GSMDevice,
    GenericDevice,
    KeyFobDevice,
    MotionDevice,
    PGMDevice,
    SmokeDevice,
)


DEVICE_TYPES = {
    "GSM": GSMDevice,
    "PGM": PGMDevice,
}

DEVICE_SUBTYPES = {
    "CONTACT": ContactDevice,
    "MC303_VANISH": ContactDevice,
    "MOTION_CAMERA": CameraDevice,
    "SMOKE": SmokeDevice,
    "BASIC_KEYFOB": KeyFobDevice,
    "KEYFOB_ARM_LED": KeyFobDevice,
    "FLAT_PIR_SMART": MotionDevice,
    "WL_SIREN": GenericDevice,
}



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyvisonicalarm",
    version="0.1.0b1",
    author="Mark Parker",
    author_email="msparker@sky.com",
    description="A simple library for the Visonic Alarm API written in Python 3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/msp1974/pyvisonicalarm",
    packages=setuptools.find_packages(),
    install_requires=["requests"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)



import json
import requests

from .const import (
    TEXT_STATUS_AWAY,
    TEXT_STATUS_DISARM,
    TEXT_STATUS_HOME,
    RequestType,
    VisonicURL,
)

from .exceptions import *


class API(object):
    """Class used for communication with the Visonic API"""

    # Client configuration
    __app_type = "com.visonic.powermaxapp"
    __user_agent = "Dart/2.10 (dart:io)"
    __rest_version = "9.0"

    # API tokens
    __user_token = None
    __session_token = None

    # Use a session to reuse one TCP connection instead of creating a new
    # connection for every call to the API
    __session = None
    __timeout = 4

    def __init__(self, hostname, app_id):
        """Class constructor initializes all URL variables."""

        # Set connection specific details
        self.__hostname = hostname
        self.__app_id = app_id

        # Create a new session
        self.__session = requests.session()

    def set_rest_version(self, version):
        """Set which version to use when connection to the API."""
        self.__rest_version = version

    def __raise_on_bad_request(self, error):
        """Raise an exception when the API returns a bad request."""
        api = json.loads(error.decode("utf-8"))
        # print(api)

        if api["error"] == 10001:  # BadRequestParams
            for pair in api["extras"]:
                if pair["value"] == "incorrect":
                    if pair["key"] == "panel_serial":
                        raise PanelSerialIncorrectError()
                    elif pair["key"] == "reset_password_code":
                        raise ResetPasswordCodeIncorrectError()
                if pair["value"] == "required":
                    if pair["key"] == "panel_serial":
                        raise PanelSerialRequiredError()
                    elif pair["key"] == "email":
                        raise EmailRequiredError()
                    elif pair["key"] == "password":
                        raise PasswordRequiredError()
                    elif pair["key"] == "app_id":
                        raise AppIDRequiredError()
                    elif pair["key"] == "user_code":
                        raise UserCodeRequiredError()
                    elif pair["key"] == "new_password":
                        raise PasswordRequiredError()
                if pair["value"] == "already_granted":
                    raise AlreadyGrantedError()
                if pair["value"] == "already_linked":
                    raise AlreadyLinkedError()
                if pair["key"] == "new_password":
                    raise NewPasswordStrengthError()
        elif api["error"] == 10004:  # WrongCombination
            for pair in api["extras"]:
                if pair["value"] == "wrong_combination":
                    if pair["key"] == "email" or pair["key"] == "password":
                        raise WrongUsernameOrPasswordError()
                    if (
                        pair["key"] == "panel_serial"
                        or pair["key"] == "master_user_code"
                    ):
                        raise WrongPanelSerialOrMasterUserCodeError()
        elif api["error"] == 10021:  # WrongUserCode
            raise UserCodeIncorrectError()
        elif api["error"] == 400 and api["error_reason_code"] == "PanelNotConnected":
            raise PanelNotConnectedError()

        # Raise a generic error when the library has no
        # specific exception implemented yet.
        raise UndefinedBadRequestError(str(api))

    def __raise_on_forbidden(self, error):
        """Raise an exception when the API returns a forbidden error."""
        api = json.loads(error.decode("utf-8"))
        # print(api)

        if api["error"] == 10010:  # NotAllowed
            raise NotAllowedError()
        elif api["error"] == 10002:  # UserAuthRequired
            raise UserAuthRequiredError()

        # Raise a generic error when the library has no
        # specific exception implemented yet.
        raise UndefinedForbiddenError(str(api))

    def __raise_on_unauthorized(self, error):
        """Raise an exception when the API returns a unauthorized error."""
        api = json.loads(error.decode("utf-8"))
        # print(api)

        # Raise an exception when we are not authorized to access the endpoint
        raise UnauthorizedError(str(api))

    def __send_request(
        self,
        endpoint,
        with_session_token=True,
        with_user_token=True,
        data_json=None,
        request_type=RequestType.GET,
    ):
        """Send a GET or POST request to the server. Includes the Session-Token
        only if with_session_token is True."""

        # Add host and api version to url
        if endpoint == VisonicURL.VERSION:
            url = f"{VisonicURL.BASE.format(self.__hostname)}/{endpoint}"
        else:
            url = f"{VisonicURL.BASE.format(self.__hostname)}/{self.__rest_version}/{endpoint}"

        # Prepare the headers to be sent
        headers = {
            "Host": self.__hostname,
            "Connection": "keep-alive",
            "Accept": "*/*",
            "User-Agent": self.__user_agent,
            "Accept-Language": "en-us",
            "Accept-Encoding": "gzip",
        }

        # Only needed for POST requests
        if request_type == "POST":
            headers["Content-Type"] = "application/json"
            headers["Content-Length"] = str(len(data_json))

        # Include the session token in the header
        if with_session_token:
            headers["Session-Token"] = self.__session_token

        # Include the user authentication token in the header
        if with_user_token:
            headers["User-Token"] = self.__user_token

        # Perform the request and raise an exception
        # if the response is not OK (HTML 200)
        try:
            if request_type == "GET":
                response = self.__session.get(
                    url, headers=headers, timeout=self.__timeout
                )
            elif request_type == "POST":
                response = self.__session.post(
                    url, headers=headers, data=data_json, timeout=self.__timeout
                )
            response.raise_for_status()
        except requests.exceptions.ConnectTimeout:
            raise ConnectionTimeoutError(
                f"Connection to '{self.__hostname}' timed out after {str(self.__timeout)} seconds."
            )
            return None
        except requests.exceptions.HTTPError as e:
            api = json.loads(response.content.decode("utf-8"))
            if "400 Client Error: Bad Request" in str(e):
                self.__raise_on_bad_request(response.content)
            elif "401 Client Error: Unauthorized" in str(e):
                self.__raise_on_unauthorized(response.content)
            elif "403 Client Error: Forbidden" in str(e):
                self.__raise_on_forbidden(response.content)
            elif "404 Client Error: Not Found" in str(e):
                raise NotFoundError()
            elif "420 Client Error:" in str(e):
                # TODO: {'error': 10020, 'error_message': 'Login temporary blocked', 'error_reason_code': 'LoginTemporaryBlocked', 'extras': [{'key': 'timeout', 'value': 44}]} // 44 = seconds to unblocked
                # print(api)
                raise LoginTemporaryBlockedError(
                    f"Login is temporary blocked due to too many failed login attempts ({api['extras'][0]['count']} seconds remaining)."
                )
            elif "440 Client Error: Session token not found" in str(e):
                raise SessionTokenError()
            elif "442 Client Error: Login attempts limit reached" in str(e):
                raise LoginAttemptsLimitReachedError("Login attempts limit reached.")
            elif "444 Client Error: Wrong user code" in str(e):
                raise InvalidUserCodeError(
                    "Authentication failed due to wrong user code."
                )
            else:
                print(api)
                raise

        # Check HTTP response code
        if response.status_code == requests.codes.ok:
            return json.loads(response.content.decode("utf-8"))
        else:
            return None

    ######################
    # Public API methods #
    ######################

    @property
    def session_token(self):
        """Property to keep track of the session token."""
        return self.__session_token

    @property
    def hostname(self):
        """Property to keep track of the API servers hostname."""
        return self.__hostname

    @property
    def user_token(self):
        """Property to keep track of the user token beeing assigned during authentication."""
        return self.__user_token

    @property
    def app_id(self):
        """Property to keep track of the user id (UUID) beeing used."""
        return self.__app_id

    def get_version_info(self):
        """Find out which REST API versions are supported."""
        return self.__send_request(
            VisonicURL.VERSION,
            with_session_token=False,
            with_user_token=False,
            request_type=RequestType.GET,
        )

    def authenticate(self, email, password):
        """Try to authenticate and get a user auth token."""
        auth_info = {
            "email": email,
            "password": password,
            "app_id": self.__app_id,
        }

        auth_json = json.dumps(auth_info, separators=(",", ":"))
        res = self.__send_request(
            VisonicURL.AUTH,
            with_session_token=False,
            with_user_token=False,
            data_json=auth_json,
            request_type=RequestType.POST,
        )
        if res is not None:
            self.__user_token = res["user_token"]
            return True
        else:
            return False

    def is_logged_in(self):
        """Check if the session token is still valid."""
        try:
            self.get_status()
            return True
        except requests.HTTPError:
            return False

    def access_grant(self, user_id, email):
        """Grant a user access to the alarm panel via the API."""
        user_data = {"user": user_id, "email": email}
        user_json = json.dumps(user_data, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.ACCESS_GRANT, data_json=user_json, request_type=RequestType.POST
        )

    def access_revoke(self, user_id):
        """Revoke access to the alarm panel via the API for a user."""
        user_data = {"user": user_id}
        user_json = json.dumps(user_data, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.ACCESS_REVOKE, data_json=user_json, request_type=RequestType.POST
        )

    def activate_siren(self):
        """Activate the siren (sound the alarm)."""
        siren_data = {}
        siren_json = json.dumps(siren_data, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.ACTIVATE_SIREN,
            data_json=siren_json,
            request_type=RequestType.POST,
        )

    def disable_siren(self, mode):
        """Disable the siren (mute the alarm)."""
        siren_data = {"mode": mode}
        siren_json = json.dumps(siren_data, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.DISABLE_SIREN,
            data_json=siren_json,
            request_type=RequestType.POST,
        )

    def get_alarms(self):
        """Get the current alarms."""
        return self.__send_request(VisonicURL.ALARMS)

    def get_alerts(self):
        """Get the current alerts."""
        return self.__send_request(VisonicURL.ALERTS)

    def get_cameras(self):
        """Get the cameras in the system."""
        return self.__send_request(VisonicURL.CAMERAS)

    def get_devices(self):
        """Get all device specific information."""
        return self.__send_request(VisonicURL.DEVICES)

    def get_email_notifications(self):
        """Get settings for the email notifications."""
        return self.__send_request(VisonicURL.NOTIFICATIONS_EMAIL)

    def get_events(self):
        """Get the alarm panel events."""
        return self.__send_request(VisonicURL.EVENTS)

    def get_feature_set(self):
        """Get the alarm panel feature set."""
        return self.__send_request(VisonicURL.FEATURE_SET)

    def get_locations(self):
        """Get all locations in the alarm system."""
        return self.__send_request(VisonicURL.LOCATIONS)

    def get_panel_info(self):
        """The general panel information is only supported in version 4.0."""
        return self.__send_request(VisonicURL.PANEL_INFO)

    def get_panels(self):
        """Get a list of panels."""
        return self.__send_request(VisonicURL.PANELS)

    def get_process_status(self, process_token):
        """Get the current status of a process running on API server."""
        return self.__send_request(VisonicURL.PROCESS_STATUS.format(process_token))

    def get_smart_devices(self):
        """Get a list of smart devices."""
        return self.__send_request(VisonicURL.SMART_DEVICES)

    def get_smart_devices_settings(self):
        """Get a list of smart devices settings."""
        return self.__send_request(VisonicURL.SMART_DEVICES_SETTINGS)

    def get_status(self):
        """Get the current status of the alarm system."""
        return self.__send_request(VisonicURL.STATUS)

    def get_troubles(self):
        """Get the current troubles."""
        return self.__send_request(VisonicURL.TROUBLES)

    def get_users(self):
        """Get information about the active users.
        Note: Only master users can see the active_user_ids!"""
        return self.__send_request(VisonicURL.USERS)

    def get_wakeup_sms(self):
        """Get the settings needed to wake up the alarm panel via SMS."""
        return self.__send_request(VisonicURL.WAKEUP_SMS)

    def panel_add(
        self, alias: str, panel_serial: str, access_proof: str, master_user_code: str
    ):
        """Add a new alarm panel to the user account. A master user code is required."""
        panel_data = {
            "alias": alias,
            "panel_serial": panel_serial,
            "access_proof": access_proof,
            "master_user_code": master_user_code,
        }
        panel_json = json.dumps(panel_data, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.PANEL_ADD, data_json=panel_json, request_type=RequestType.POST
        )

    def panel_login(self, panel_serial: str, user_code: str):
        """Try to login to the alarm panel and get a session token."""
        login_info = {
            "user_code": user_code,
            "app_type": self.__app_type,
            "app_id": self.__app_id,
            "panel_serial": panel_serial,
        }

        login_json = json.dumps(login_info, separators=(",", ":"))
        res = self.__send_request(
            VisonicURL.PANEL_LOGIN,
            with_session_token=False,
            data_json=login_json,
            request_type=RequestType.POST,
        )
        if res:
            self.__session_token = res["session_token"]
            return True
        else:
            return False

    def panel_rename(self, alias: str, panel_serial: str):
        """Rename an alarm panel."""
        panel_data = {
            "panel_serial": panel_serial,
            "alias": alias,
        }
        panel_json = json.dumps(panel_data, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.PANEL_RENAME, data_json=panel_json, request_type=RequestType.POST
        )

    def panel_unlink(self, panel_serial: str, password: str, app_id: str):
        """Unlink an alarm panel from the user account."""
        panel_data = {
            "panel_serial": panel_serial,
            "password": password,
            "app_id": app_id,
        }
        panel_json = json.dumps(panel_data, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.PANEL_UNLINK, data_json=panel_json, request_type=RequestType.POST
        )

    def password_reset(self, email: str):
        """Request a password reset email. An email will be sent to the email address provided."""
        reset_data = {"email": email}
        reset_json = json.dumps(reset_data, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.PASSWORD_RESET,
            data_json=reset_json,
            request_type=RequestType.POST,
        )

    def password_reset_complete(self, reset_password_code: str, new_password: str):
        """Complete the password reset request."""
        reset_data = {
            "reset_password_code": reset_password_code,
            "new_password": new_password,
            "app_id": self.__app_id,
        }
        reset_json = json.dumps(reset_data, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.PASSWORD_RESET_COMPLETE,
            data_json=reset_json,
            request_type=RequestType.POST,
        )

    def set_email_notifications(self, mode: str):
        """Set settings for the email notifications."""
        notification_data = {"mode": mode}
        notification_json = json.dumps(notification_data, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.NOTIFICATIONS_EMAIL,
            data_json=notification_json,
            request_type=RequestType.POST,
        )

    def set_bypass_zone(self, zone: int, set_enabled: bool):
        """Enable or disable bypass mode for a zone."""
        bypass_data = {"zone": zone, "set": set_enabled}
        bypass_json = json.dumps(bypass_data, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.SET_BYPASS_ZONE,
            data_json=bypass_json,
            request_type=RequestType.POST,
        )

    def set_name(self, object_class: str, id: int, name: str):
        """Set the name of any type of object in the alarm system."""
        name_data = {"class": object_class, "id": id, "name": name}
        name_json = json.dumps(name_data, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.SET_NAME, data_json=name_json, request_type=RequestType.POST
        )

    def set_user_code(self, user_code: str, user_id: str):
        """Set the code of a user in the alarm system."""
        code_data = {"user_code": user_code, "user_id": user_id}
        code_json = json.dumps(code_data, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.SET_USER_CODE, data_json=code_json, request_type=RequestType.POST
        )

    def arm_home(self, partition: int):
        """Arm in Home mode."""
        arm_info = {"partition": partition, "state": TEXT_STATUS_HOME}
        arm_json = json.dumps(arm_info, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.SET_STATE, data_json=arm_json, request_type=RequestType.POST
        )

    def arm_away(self, partition: id):
        """Arm in Away mode."""
        arm_info = {"partition": partition, "state": TEXT_STATUS_AWAY}
        arm_json = json.dumps(arm_info, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.SET_STATE, data_json=arm_json, request_type=RequestType.POST
        )

    def disarm(self, partition: id):
        """Disarm the alarm system."""
        disarm_info = {"partition": partition, "state": TEXT_STATUS_DISARM}
        disarm_json = json.dumps(disarm_info, separators=(",", ":"))
        return self.__send_request(
            VisonicURL.SET_STATE, data_json=disarm_json, request_type=RequestType.POST
        )

    def send_get(self, url):
        """Send a custom POST request."""
        return self.__send_request(url)

    def send_post(self, url, data):
        """Send a custom POST request."""
        data_json = json.dumps(data, separators=(",", ":"))
        return self.__send_request(
            url, data_json=data_json, request_type=RequestType.POST
        )
